;* ************************************************************************ *
;* DS1990 battery-less emulator                                             *
;* ************************************************************************ *
.include "tn13Adef.inc"

;  : RSTDISBL=BODLEVEL1=WDTON=SUT1=SUT0=CKSEL0=0
; fast rising power, BOD=2.5V, 9.6MHz

.set    VERSION = 0x0100
.equ    XTAL    = 9600000
.equ    XTK     = int(XTAL/1000)

;**** Global Register Variables ****
;.undef  YL
;.undef  YH
;.undef  ZL
;.undef  ZH

;   NVM
.equ    MA_KEYN     = 0x00                  ;    

; 
.equ    IBRST_TM    = 90                    ;     
.equ    IBHAR_TM    = 32                    ;     PRECENSE_PULSE
.equ    IBPRC_TM    = 120                   ;  PRECENSE_PULSE
.equ    IBRZE_TM    = 30                    ;   '0'
.equ    IBWZE_TM    = 40                    ;   '0'
.equ    IBWON_TM    = 5                     ;   '1' ( )


; *    *
;  B
.equ    CTBout  = PORTB                     ;  
.equ    CTBcfg  = DDRB                      ;   
.equ    CTBin   = PINB                      ;  

.equ    DSpin   = 4                         ; / 1-Wire

.equ    CTB_CFG = (0<<DSpin)                ;  cfg 
.equ    CTB_OUT = (0<<DSpin)                ;  out 

; *   *
.def    tmp1    = R16                       ;   
.def    tmp2    = R17                       ;   
.def    tmp3    = R18                       ;   

.equ    bDSN    = SRAM_START                ;   SN  RAM, 8 

;
;  
        .cseg                               ; CODE segment
;______________________________________________________________
;  
        .org    0
        rjmp    main_rstcold                ;  

;______________________________________________________________
;     ; tmp2 - , tmp1 - 
eewr:
        sbic    EECR, EEWE                  ; test for previous operation end
        rjmp    eewr                        ; close wait loop
        out     EEAR, tmp2                  ;  
        out     EEDR, tmp1                  ;  
        sbi     EECR, EEMWE                 ;   
        sbi     EECR, EEWE                  ;   
        ret                                 ; 
;______________________________________________________________
;     ; tmp2 - , tmp1 - 
eerd:
        sbic    EECR, EEWE                  ; test for previous operation end
        rjmp    eerd                        ; close wait loop
        out     EEAR, tmp2                  ;  
        sbi     EECR, EERE                  ;  
        in      tmp1, EEDR                  ;  
        ret                                 ; 

;______________________________________________________________
ibcrc8:
;    CRC8
; IN  - tmp1, tmp2(CRC8)
; OUT - tmp2(CRC8)
; CHG - tmp1, tmp2, tmp3
        ldi     tmp3, 8                     ; -   
ibcr_crclp:
        push    tmp1                        ;    
        eor     tmp1, tmp2                  ;  . 
        lsr     tmp2                        ;  CRC8
        sbrs    tmp1, 0                     ;  . 
        rjmp    ibcr_inror                  ; ,  . ==0
        ldi     tmp1, 0x8C                  ; crc ^= 0x8C
        eor     tmp2, tmp1                  ;
ibcr_inror:
        pop     tmp1                        ;  .
        lsr     tmp1                        ;    .
        dec     tmp3                        ;   
        brne    ibcr_crclp                  ;  
        ret                                 ; 
;______________________________________________________________
iwaithi:
;  Hi-   RST
; IN  - 
; OUT - Z=1,  RST
; CHG - X
.equ    WHRST_CN = int(IBRST_TM*XTK/1000/7) ;    > 60
        ldi     XL, low(WHRST_CN)           ;  
        ldi     XH, high(WHRST_CN)          ;
iwhi_wthilp:
        wdr                                 ;~1,  WDT
        sbic    CTBin, DSpin                ;~1/2, ,   Lo
        rjmp    iwhi_bushi                  ; ,    ( )
        sbiw    XH:XL, 1                    ;~2,  
        brne    iwhi_wthilp                 ;~1/2,     
        sez                                 ;   RST
        ret                                 ; 
iwhi_bushi:
        clz                                 ;   RST
        ret                                 ; 
;______________________________________________________________
iwaitrst:
;  ,   Lo > IBRST_TM
; IN  - 
; OUT - 
; CHG - X
.equ    WTRST_CN = int(IBRST_TM*XTK/1000/5) ;     450
        ldi     XL, low(WTRST_CN)           ;     Hi
        ldi     XH, high(WTRST_CN)          ;
irst_hldlp:
        wdr                                 ;~1,  WDT
        sbic    CTBin, DSpin                ;~1/2, ,   
        rjmp    iwaitrst                    ;~2, ,     (Hi)
        sbiw    XH:XL, 1                    ;~2,  
        brne    irst_hldlp                  ;~1/2, ,  rst
irst_wthilp:
        rcall   iwaithi                     ;   
        breq    irst_wthilp                 ; ,    (Lo)
        ret                                 ; 

;______________________________________________________________
iprecense:
;       
; IN  - 
; OUT - Z=1,  RST
; CHG - X
.equ    WPRHAR_CN = int(IBHAR_TM*XTK/1000/5)    ;  HOLD_AFTER_RESET
.equ    WPRPUL_CN = int(IBPRC_TM*XTK/1000/5)    ;  PRESENE_PULSE
        rcall   iwaithi                     ;  Hi
        breq    iprc_ret                    ; ,   
        ldi     XL, low(WPRHAR_CN)          ;     
        ldi     XH, high(WPRHAR_CN)         ;
iprc_hldlp:
        wdr                                 ;~1,  WDT
        sbiw    XH:XL, 1                    ;~2,  
        brne    iprc_hldlp                  ;~1/2,   
;   
        sbi     CTBcfg, DSpin               ;  Lo  :  PRESENE_PULSE
        ldi     XL, low(WPRPUL_CN)          ;~1,   PRESENE_PULSE
        ldi     XH, high(WPRPUL_CN)         ;~1
iprc_prclp:
        wdr                                 ;~1,  WDT
        sbiw    XH:XL, 1                    ;~2,   
        brne    iprc_prclp                  ;~1/2,   
        cbi     CTBcfg, DSpin               ;  :  PRESENE_PULSE
        clz                                 ;   
iprc_ret:
        ret                                 ; 

;______________________________________________________________
ibreadbit:
;  1    1-Wire
; IN  - 
; OUT -  NZ, tmp1.7
;        Z,  RST
; CHG - tmp1,tmp3,X
        ldi     tmp3, 1                     ; -  
        rjmp    ibrd_bit                    ;    
;______________________________________________________________
ibread:
;  8-   1-Wire
; IN  - 
; OUT -  NZ, tmp1=<RxD>
;        Z,  RST
; CHG - tmp1,tmp3,X
.equ    RBRST_CN = int(IBRST_TM*XTK/1000/7)         ;   
.equ    RBBIT_CN = RBRST_CN-int(IBRZE_TM*XTK/1000/7);   0
        ldi     tmp3, 8                     ;~1, -  
ibrd_bit:
        rcall   iwaithi                     ;  Hi
        breq    ibrd_ret                    ; ,   
;   
ibrd_bitlp:
        wdr                                 ;~1,  WDT
        sbic    CTBin,DSpin                 ;~1/2, ,  ,   
        rjmp    ibrd_bitlp                  ;~2,    
;      
        ldi     XL, low(RBRST_CN)           ;     
        ldi     XH, high(RBRST_CN)          ;
ibrd_riselp:
        wdr                                 ;~1,  WDT
        sbiw    XH:XL, 1                    ;~2,  
        breq    ibrd_ret                    ;~1/2, ,  rst
        sbis    CTBin, DSpin                ; ,  
        rjmp    ibrd_riselp                 ;     
;  0  1
        subi    XL, low(RBBIT_CN)           ;   0  1
        sbci    XH, high(RBBIT_CN)          ;
        ror     tmp1                        ;     
        dec     tmp3                        ;   
        brne    ibrd_bitlp                  ;    
        com     tmp1                        ;  
        clz                                 ;   
ibrd_ret:
        ret                                 ; 

;______________________________________________________________
ibwritebit:
;  1    1-Wire
; IN  - tmp1.0
; OUT -  Z=1,  RST
; CHG - tmp1,tmp3,X
        ldi     tmp3, 1                     ; -  
        rjmp    ibwr_bit                    ;    
;______________________________________________________________
ibwrite:
;  8-   1-Wire
; IN  - tmp1
; OUT -  Z=1,  RST
; CHG - tmp1,tmp3,X
.equ    WBRST_CN = int(IBRST_TM*XTK/1000/7) ;   
.equ    WBZBIT_CN = int(IBWZE_TM*XTK/1000/5);   '0'
.equ    WBOBIT_CN = int(IBWON_TM*XTK/1000/5);   '1'
        ldi     tmp3, 8                     ; -  
;   
ibwr_bit:
        rcall   iwaithi                     ;  Hi
        breq    ibwr_ret                    ; ,  
        ror     tmp1                        ;      
ibwr_waitlp:
        wdr                                 ;  WDT
        sbic    CTBin,DSpin                 ; ,  
        rjmp    ibwr_waitlp                 ;      
;   1  0
        ldi     XL, low(WBOBIT_CN)          ;    '1'
        ldi     XH, high(WBOBIT_CN)         ;
        brcs    ibwr_holdlp                 ; ,    '1'
        sbi     CTBcfg, DSpin               ;  Lo  ,   '0'
        ldi     XL, low(WBZBIT_CN)          ;   Lo-   '0'
        ldi     XH, high(WBZBIT_CN)         ;
ibwr_holdlp:
        wdr                                 ;~1,  WDT
        sbiw    XH:XL, 1                    ;~2,  
        brne    ibwr_holdlp                 ;~1/2,   
        cbi     CTBcfg, DSpin               ;  
        dec     tmp3                        ;   
        brne    ibwr_bit                    ;    
        clz                                 ;   
ibwr_ret:
        ret                                 ; 

;______________________________________________________________
icm_search:
;   SEARCH_ROM
; IN  - 
; OUT -  Z=1,  RST
;            Z=0 & C=0,    DS_SN, WAIT_RST
;            Z=0 & C=1,    DS_SN
; CHG - tmp1,tmp2,tmp3,X,Z
        ldi     ZL, low(bDSN)               ;   SN + CRC8 (8)
        ldi     ZH, high(bDSN)              ;
icsr_bytelp:
        ldi     tmp2, 8                     ; -   
        ld      tmp1, Z+                    ;    SN  ROM  tmp1
icsr_bitlp:
        push    tmp1                        ;   SN
        rcall   ibwritebit                  ;  tmp1.0
        pop     tmp1                        ;   SN
        breq    icsr_rst                    ; ,   
        push    tmp1                        ;   SN
        com     tmp1                        ;   tmp1.0
        rcall   ibwritebit                  ;  tmp1.0
        pop     tmp1                        ;   SN
        breq    icsr_rst                    ; ,   
        push    tmp1                        ;   SN
        rcall   ibreadbit                   ;     
        pop     tmp3                        ;   SN  tmp3
        breq    icsr_rst                    ; ,   
        rol     tmp1                        ;  b.7->b.0, ..     Tx/Rx
        rol     tmp1                        ;
        eor     tmp1, tmp3                  ;   bDSN.n   
        clc                                 ;    DS_SN
        sbrc    tmp1, 0                     ;  .,   
        rjmp    icsr_nsel                   ; ,   ,  
        mov     tmp1, tmp3                  ;   SN
        lsr     tmp1                        ;     
        dec     tmp2                        ;   
        brne    icsr_bitlp                  ;   
        cpi     ZL, low(bDSN + 8)           ;   SN
        brne    icsr_bytelp                 ;   
        sec                                 ;    DS_SN
icsr_nsel:
        clz                                 ;   RST
icsr_rst:
        ret                                 ; 
;______________________________________________________________
icm_readrom:
;   READ_ROM
; IN  - 
; OUT -  Z=1,  RST
;            Z=0,  WAIT_RST
; CHG - tmp1,tmp3,X,Z
        ldi     ZL, low(bDSN)               ;   SN + CRC8 (8)
        ldi     ZH, high(bDSN)              ;
icrr_lp:
        ld      tmp1, Z+                    ;    SN  ROM  tmp1
        rcall   ibwrite                     ;    
        breq    icsr_rst                    ; ,   RST
        cpi     ZL, low(bDSN + 8)           ;   
        brne    icrr_lp                     ;  
        rjmp    icsr_nsel                   ;    DS_SN
;______________________________________________________________
icm_matchrom:
;   MATCH_ROM
; IN  - 
; OUT -  Z=1,  RST
;            Z=0 & C=0,    DS_SN, WAIT_RST
;            Z=0 & C=1,    DS_SN
; CHG - tmp1,tmp3,X,Z
        ldi     ZL, low(bDSN)               ;   SN + CRC8 (8)
        ldi     ZH, high(bDSN)              ;
icmr_lp:
        rcall   ibread                      ;    
        breq    icmr_ret                    ; ,  RST
        ld      tmp3, Z+                    ;    RAM-
        cp      tmp1, tmp3                  ;       SN
        clc                                 ;    DS_SN
        brne    icmr_nsel                   ; ,     DS_SN
        cpi     ZL, low(bDSN + 8)           ;   
        brne    icmr_lp                     ;    
        sec                                 ;    DS_SN
icmr_nsel:
        clz                                 ;   RST
icmr_ret:
        ret                                 ; 



;______________________________________________________________
;  , power-off   
pwroff:
        ldi     tmp1, (1<<SE) | (1<<SM1)    ;  POWER_DOWN   sleep
        out     MCUCR, tmp1                 ;
        ldi     tmp1, (1<<BPDS) | (1<<BPDSE);    BOD
        out     BODCR, tmp1                 ;   BOD,  3    sleep
        cbr     tmp1,(1<<BPDSE)
        out     BODCR, tmp1                 ;  BOD
        sleep                               ;  
;______________________________________________________________
;  ,     
main_rstcold:
        in      tmp3, MCUSR                 ;   
        sbi     ACSR, ACD                   ;  
        ldi     tmp1, (1<< PRTIM0) | (1<<PRADC) ;     ADC  TCNT0
        out     PRR, tmp1                   ;  TCNT0 & ADC
;   
        cli                                 ;   
        wdr                                 ;   
        in      tmp1, WDTCR                 ;   
        ori     tmp1, (1<<WDCE) | (1<<WDE)  ;   WD
        out     WDTCR, tmp1                 ;    , 4   
        ldi     tmp1,(1<<WDE) | (1<<WDP2) | (0<<WDP1) | (0<<WDP0)   ;  WD  0.25
        out     WDTCR, tmp1                 ;    
;  
        ldi     tmp1, low(RAMEND)           ;   
        out     SPL, tmp1                   ;  
;     B
        ldi     tmp1, CTB_OUT               ;   
        out     CTBout, tmp1                ;     
        ldi     tmp1, CTB_CFG               ;    B
        out     CTBcfg, tmp1                ;  I/O 
;     NVM
;   
        sbrs    tmp3, WDRF                  ;    WD 
        rjmp    pwroff                      ;   POFF ,  1 
;      
        ldi     tmp2, MA_KEYN               ;   NVM
        rcall   eerd                        ;  NVM
        inc     tmp1                        ;   
        rcall   eewr                        ;    NVM
        andi    tmp1, 3                     ;   :  tmp1    0..3
        ldi     XL, low(bDSN)               ;   RAM-
        ldi     XH, high(bDSN)              ;
        ldi     ZL, low(DS_SN << 1)         ;    
        ldi     ZH, high(DS_SN << 1)        ;
        swap    tmp1                        ; *16
        lsr     tmp1                        ; /2
        add     ZL, tmp1                    ;    ROM:
        clr     tmp1                        ; Z = DS_SN + 8 * tmp1
        adc     ZH, tmp1                    ;
        clr     tmp2                        ; CRC8 = 0
main_setdsnlp:
        lpm     tmp1, Z+                    ;    ROM
        st      X+, tmp1                    ;    RAM
        rcall   ibcrc8                      ;  CRC8  tmp2
        cpi     XL, low(bDSN + 7)           ;   
        brne    main_setdsnlp               ;    
        st      X, tmp2                     ;  SN  RAM CRC8
;   
        sei                                 ;  
        rjmp    main_precense               ;    

;   
main_waitrst:
        rcall   iwaitrst                    ;   ,    Lo
main_precense:
        rcall   iprecense                   ;   
        breq    main_precense               ; ,     
main_romlp:
        rcall   ibread                      ;   
        breq    main_precense               ; ,   
        cpi     tmp1, 0x0F                  ;   0F - READ_ROM
        breq    main_cmd33                  ;  READ_ROM,   
        cpi     tmp1, 0x33                  ;   33 - READ_ROM
        breq    main_cmd33                  ;  READ_ROM,   
        cpi     tmp1, 0x55                  ;   55 - READ_ROM
        breq    main_cmd55                  ;  MATCH_ROM,   
        cpi     tmp1, 0xCC                  ;   CC - SKIP_ROM
        breq    main_romlp                  ;  SKIP_ROM,   
        cpi     tmp1, 0xF0                  ;   F0 - SEARCH_ROM
        breq    main_cmdF0                  ;  SEARCH_ROM,   
        rjmp    main_waitrst                ;    ,   

; [0F][33] READ_ROM
main_cmd33:
        rcall   icm_readrom                 ;  
        breq    main_precense               ; ,   
        rjmp    main_waitrst                ;      
; [0x55] MATCH_ROM
main_cmd55:
        rcall   icm_matchrom                ;  
        breq    main_precense               ; ,   
        rjmp    main_waitrst                ;    
; [F0] SEARCH_ROM
main_cmdF0:
        rcall   icm_search                  ;  
        breq    main_precense               ; ,   
        rjmp    main_waitrst                ;    
;______________________________________________________________
DS_SN:
; SN             FAM   SN0   SN1   SN2   SN3   SN4   SN5   CRC
        .db     0x01, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF  ; K1
        .db     0x01, 0x00, 0x00, 0x00, 0x00, 0x90, 0x19, 0xFF  ; K2
        .db     0x01, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0xFF  ; K3
        .db     0x01, 0x78, 0x00, 0x48, 0xFD, 0xFF, 0xFF, 0xFF  ; K4
